"""
Robust piecewise regression.
"""

# Guillermo Navas-Palencia <g.navas.palencia@gmail.com>
# Copyright (C) 2020

import numbers

import numpy as np

from sklearn.base import BaseEstimator
from sklearn.utils import check_array
from sklearn.utils import check_consistent_length
from sklearn.exceptions import NotFittedError

from .direct import lsq_direct
from .direct import lsq_direct_separated
from .cvx_socp import socp
from .cvx_socp import socp_separated
from .cvx_qp import qp
from .cvx_qp import qp_separated


def _check_parameters(objective, regularization, degree, continuous,
                      monotonic_trend, solver, h_epsilon, quantile, reg_l1,
                      reg_l2, verbose):

    if objective not in ("l1", "l2", "huber", "quantile"):
        raise ValueError('Invalid value for objective. Allowed string '
                         'values are "l1", "l2", "huber" and "quantile".')

    if regularization is not None:
        if regularization not in ("l1", "l2"):
            raise ValueError('Invalid value for regularization. Allowed '
                             'string values are "l1" and "l2".')

    if not isinstance(degree, numbers.Integral) or not 0 <= degree <= 5:
        raise ValueError("degree must be an integer in [0, 5].")

    if not isinstance(continuous, bool):
        raise TypeError("continuous must be a boolean; got {}."
                        .format(verbose))

    if monotonic_trend is not None:
        if monotonic_trend not in ("ascending", "descending", "convex",
                                   "concave", "peak", "valley"):
            raise ValueError('Invalid value for monotonic trend. Allowed '
                             'string values are "ascending", "descending", '
                             '"convex", "concave", "peak" and "valley".')

        if (monotonic_trend in ("convex", "concave", "peak", "valley")
                and not continuous):
            raise ValueError('Option monotonic trend "convex", "concave", '
                             '"peak" and "valley" valid if continuous=True.')

    if solver not in ("auto", "ecos", "osqp", "direct", "scs", "highs"):
        raise ValueError('Invalid value for solver. Allowed string '
                         'values are "auto", "ecos", "osqp", "direct", '
                         '"scs" and "highs".')

    if not isinstance(h_epsilon, numbers.Number) or h_epsilon < 1.0:
        raise ValueError("h_epsilon must a number >= 1.0; got {}."
                         .format(h_epsilon))

    if not isinstance(quantile, numbers.Number) or not 0.0 < quantile < 1.0:
        raise ValueError("quantile must be a value in (0, 1); got {}."
                         .format(quantile))

    if not isinstance(reg_l1, numbers.Number) or reg_l1 < 0.0:
        raise ValueError("reg_l1 must be a positive value; got {}."
                         .format(reg_l1))

    if not isinstance(reg_l2, numbers.Number) or reg_l2 < 0.0:
        raise ValueError("reg_l2 must be a positive value; got {}."
                         .format(reg_l2))

    if not isinstance(verbose, bool):
        raise TypeError("verbose must be a boolean; got {}.".format(verbose))


def _choose_method(objective, degree, continuous, monotonic_trend, solver,
                   bounded, regularization):

    if solver == "auto":
        if bounded or regularization is not None:
            if continuous:
                return "socp"
            else:
                return "socp_separated"
        else:
            if objective == "l2":
                if monotonic_trend is None:
                    if continuous:
                        return "lsq_direct"
                    else:
                        return "lsq_direct_separated"
                elif degree <= 1:
                    if continuous or degree == 0:
                        return "qp"
                    else:
                        return "qp_separated"
                else:
                    if continuous:
                        return "socp"
                    else:
                        return "socp_separated"
            else:
                if continuous:
                    return "socp"
                else:
                    return "socp_separated"
    elif solver == "direct":
        if (objective != "l2" or monotonic_trend is not None or bounded
                or regularization is not None):
            raise ValueError('solver "direct" only for objective="l2", '
                             'monotonic_trend=None, lb=ub=None, and '
                             'regularization=None.')
        else:
            if continuous:
                return "lsq_direct"
            else:
                return "lsq_direct_separated"
    elif solver == "osqp":
        if objective == "l2" and regularization is None:
            if continuous:
                return "qp"
            else:
                return "qp_separated"
        else:
            raise ValueError('solver="osqp" only for objective="l2" and '
                             'regularization=None.')
    elif solver in ("ecos", "scs"):
        if continuous or degree == 0:
            return "socp"
        else:
            return "socp_separated"
    elif solver == "highs":
        if objective == "l1" and regularization in (None, "l1"):
            if continuous:
                return "socp"
            else:
                return "socp_separated"
        else:
            raise ValueError('solver="highs" only for objective="l1" and '
                             'regularization in (None, "l1").')


def _check_bounds(lb, ub):
    if lb is not None:
        if not isinstance(lb, numbers.Number):
            raise TypeError("lb must be a number or None; got {}."
                            .format(type(bool)))

    if ub is not None:
        if not isinstance(ub, numbers.Number):
            raise TypeError("ub must be a number or None; got {}."
                            .format(type(bool)))

    if lb is not None and ub is not None:
        if lb > ub:
            raise ValueError("lb must be <= ub.")


def _check_splits(splits, monotonic_trend):
    if not isinstance(splits, (list, np.ndarray)):
        raise TypeError("splits must be a list or numpy.ndarray.")

    if not len(splits):
        if monotonic_trend in ("peak", "valley"):
            raise ValueError('monotonic trend "peak" and "valley" require a '
                             'list of splits.')

        return splits
    else:
        user_splits = check_array(splits, ensure_2d=False,
                                  force_all_finite=True)

        if len(set(user_splits)) != len(user_splits):
            raise ValueError("splits are not unique.")

        sorted_idx = np.argsort(user_splits)
        user_splits = user_splits[sorted_idx]

        return user_splits


class RobustPWRegression(BaseEstimator):
    """Robust piecewise regression.

    Parameters
    ----------
    objective : str, optional (default="l2")
        The objective function. Supported objectives are "l2", "l1", "huber"
        and "quantile". Note that "l1", "huber" and "quantile" are robust
        objective functions.

    regularization: str or None (default=None)
        Type of regularization. Supported regularization are "l1" (Lasso) and
        "l2" (Ridge). If None, no regularization is applied.

    degree : int (default=1)
        The degree of the polynomials.

        * degree = 0: piecewise constant functions.
        * degree = 1: piecewise linear functions.
        * degree > 1: piecewise polynomial functions.

    continuous : bool (default=True)
        Whether to fit a continuous or discontinuous piecewise regression.

    monotonic_trend : str or None, optional (default=None)
        The monotonic trend. Supported trends are "ascending", "descending",
        "convex" and "concave". If None, then the monotonic constraint is
        disabled.

    solver : str, optional (default="auto")
        The optimizer to solve the underlying mathematical optimization
        problem. Supported solvers are `"ecos"
        <https://github.com/embotech/ecos>`_, `"osqp"
        <https://github.com/oxfordcontrol/osqp>`_, "direct", to choose the
        direct solver, and "auto", to choose the most appropriate solver for
        the problem. Version 0.3.0 added support to solvers
        `"scs" <https://github.com/cvxgrp/scs>`_ and `"highs"
        <https://github.com/ERGO-Code/HiGHS>`_.

    h_epsilon: float (default=1.35)
        The parameter h_epsilon used when ``objective="huber"``, controls the
        number of samples that should be classified as outliers.

    quantile : float (default=0.5)
        The parameter quantile is the q-th quantile to be used when
        ``objective="quantile"``.

    reg_l1 : float (default=1.0)
        L1 regularization term. Increasing this value will smooth the
        regression model. Only applicable if ``regularization="l1"``.

    reg_l2 : float (default=1.0)
        L2 regularization term. Increasing this value will smooth the
        regression model. Only applicable if ``regularization="l2"``.

    verbose : bool (default=False)
        Enable verbose output.

    Attributes
    ----------
    coef_ : numpy.ndarray of shape (n_splits + 1, degree + 1)
        Coefficients for each bin. Number of bins = n_splits + 1.
    """
    def __init__(self, objective="l2", degree=1, continuous=True,
                 monotonic_trend=None, solver="auto", h_epsilon=1.35,
                 quantile=0.5, regularization=None, reg_l1=1.0, reg_l2=1.0,
                 verbose=False):

        self.objective = objective
        self.degree = degree
        self.continuous = continuous
        self.monotonic_trend = monotonic_trend
        self.solver = solver
        self.h_epsilon = h_epsilon
        self.quantile = quantile
        self.regularization = regularization
        self.reg_l1 = reg_l1
        self.reg_l2 = reg_l2
        self.verbose = verbose

        self.coef_ = None

        self._is_fitted = False

    def fit(self, x, y, splits, lb=None, ub=None):
        """Fit the piecewise regression according to the given training data.

        Parameters
        ----------
        x : array-like, shape = (n_samples,)
            Training vector, where n_samples is the number of samples.

        y : array-like, shape = (n_samples,)
            Target vector relative to x.

        lb : float or None (default=None)
            Add constraints to avoid values below the lower bound lb.

        ub : float or None (default=None)
            Add constraints to avoid values above the upper bound ub.

        Returns
        -------
        self : object
            Fitted piecewise regression.
        """
        _check_parameters(**self.get_params())

        # Check inputs x and y
        x = check_array(x, ensure_2d=False, force_all_finite=True)
        y = check_array(y, ensure_2d=False, force_all_finite=True)
        check_consistent_length(x, y)

        idx = np.argsort(x)
        xs = x[idx]
        ys = y[idx]

        # Check user splits
        _check_splits(splits, self.monotonic_trend)

        # Check bounds
        bounded = (lb is not None or ub is not None)

        if bounded:
            _check_bounds(lb, ub)

        # Choose the most appropriate method/solver given the parameters
        _method = _choose_method(self.objective, self.degree, self.continuous,
                                 self.monotonic_trend, self.solver, bounded,
                                 self.regularization)

        if _method == "lsq_direct":
            c, info = lsq_direct(xs, ys, splits, self.degree)
        elif _method == "lsq_direct_separated":
            c, info = lsq_direct_separated(xs, ys, splits, self.degree)
        elif _method == "socp":
            c, info = socp(xs, ys, splits, self.degree, self.continuous, lb,
                           ub, self.objective, self.monotonic_trend,
                           self.h_epsilon, self.quantile, self.regularization,
                           self.reg_l1, self.reg_l2, self.solver, self.verbose)
        elif _method == "socp_separated":
            c, info = socp_separated(xs, ys, splits, self.degree, lb, ub,
                                     self.objective, self.monotonic_trend,
                                     self.h_epsilon, self.quantile,
                                     self.solver, self.verbose)
        elif _method == "qp":
            c, info = qp(xs, ys, splits, self.degree, self.continuous, lb, ub,
                         self.monotonic_trend, self.verbose)
        elif _method == "qp_separated":
            c, info = qp_separated(xs, ys, splits, self.degree, lb, ub,
                                   self.monotonic_trend, self.verbose)

        self.coef_ = c

        if self.continuous or self.degree == 0:
            self._status = info["status"]
            self._stats = info["stats"]
        else:
            self._status = [_info["status"] for _info in info]
            self._stats = [_info["stats"] for _info in info]

        self._splits = splits
        self._is_fitted = True

        return self

    def fit_predict(self, x, y, splits, lb=None, ub=None):
        """Fit the piecewise regression according to the given training data,
        then predict.

        Parameters
        ----------
        x : array-like, shape = (n_samples,)
            Training vector, where n_samples is the number of samples.

        y : array-like, shape = (n_samples,)
            Target vector relative to x.

        lb : float or None (default=None)
            Fit impose constraints to satisfy that values are greater or equal
            than lb. In predict, values below the lower bound lb are clipped to
            lb.

        ub : float or None (default=None)
            Fit impose constraints to satisfy that values are less or equal
            than ub. In predict, values above the upper bound ub are clipped to
            ub.

        Returns
        -------
        p : numpy array, shape = (n_samples,)
            Predicted array.
        """
        return self.fit(x, y, splits, lb, ub).predict(x, lb, ub)

    def predict(self, x, lb=None, ub=None):
        """Predict using the piecewise regression.

        Parameters
        ----------
        x : array-like, shape = (n_samples,)
            Training vector, where n_samples is the number of samples.

        lb : float or None (default=None)
            Values below the lower bound lb are clipped to lb.

        ub : float or None (default=None)
            Values above the upper bound ub are clipped to ub.

        Returns
        -------
        p : numpy array, shape = (n_samples,)
            Predicted array.
        """
        self._check_is_fitted()

        bounded = (lb is not None or ub is not None)

        if bounded:
            _check_bounds(lb, ub)

        n_bins = len(self._splits) + 1
        indices = np.digitize(x, self._splits, right=False)

        pred = np.zeros(x.shape)
        for i in range(n_bins):
            mask = (indices == i)
            pred[mask] = np.polyval(self.coef_[i, :][::-1], x[mask])

        bounded = (lb is not None or ub is not None)
        if bounded:
            pred = np.clip(pred, lb, ub)

        return pred

    def _check_is_fitted(self):
        if not self._is_fitted:
            raise NotFittedError("This {} instance is not fitted yet. Call "
                                 "'fit' with appropriate arguments."
                                 .format(self.__class__.__name__))

    @property
    def status(self):
        """The status of the underlying optimization solver.

        Returns
        -------
        status : str
        """
        self._check_is_fitted()

        return self._status

    @property
    def stats(self):
        """The number of variables and constraints of the underlying
        optimization problem.

        Returns
        -------
        stats : dict
        """
        self._check_is_fitted()

        return self._stats



"""
General Second-order Cone Programming (SOCP) formulation for piecewise
regression.
"""

# Guillermo Navas-Palencia <g.navas.palencia@gmail.com>
# Copyright (C) 2020

import cvxpy as cp
import numpy as np

from .cvx import compute_change_point
from .cvx import monotonic_trend_constraints
from .cvx import problem_info
from .matrices import matrix_A
from .matrices import matrix_A_D
from .matrices import matrix_A_H
from .matrices import matrix_D
from .matrices import matrix_H
from .matrices import matrix_S
from .matrices import submatrix_A
from .matrices import submatrix_A_D
from .matrices import submatrix_D


def _model_objective(A, c, y, objective, regularization, h_epsilon, quantile,
                     reg_l1, reg_l2):
    if objective == "l1":
        obj = cp.norm(A @ c - y, 1)
    elif objective == "l2":
        obj = cp.norm(A @ c - y, 2)
    elif objective == "huber":
        obj = cp.sum(cp.huber(A @ c - y, h_epsilon))
    elif objective == "quantile":
        obj1 = 0.5 * cp.norm(A @ c - y, 1)
        obj2 = (quantile - 0.5) * cp.sum(A @ c - y)
        obj = obj1 + obj2

    if regularization == "l1":
        obj += reg_l1 * cp.norm(c, 1)
    elif regularization == "l2":
        obj += reg_l2 * cp.norm(c, 2)

    return cp.Minimize(obj)


def socp(x, y, splits, degree, continuous, lb, ub, objective, monotonic_trend,
         h_epsilon, quantile, regularization, reg_l1, reg_l2, solver, verbose):

    # Parameters
    n_bins = len(splits) + 1
    order = degree + 1

    t = None
    if monotonic_trend in ("ascending", "descending", "peak", "valley"):
        if order <= 2:
            A = matrix_A(x, splits, order)
            D = matrix_D(x, splits, order)
        else:
            A, D = matrix_A_D(x, splits, order)

        if monotonic_trend in ("peak", "valley"):
            t = compute_change_point(x, y, splits, order, monotonic_trend)
    elif monotonic_trend in ("convex", "concave"):
        if order <= 2:
            A = matrix_A(x, splits, order)
            D = matrix_H(x, splits, order)
        else:
            A, D = matrix_A_H(x, splits, order)
    else:
        A = matrix_A(x, splits, order)
        D = None

    # Decision variables
    nvar = order * n_bins
    c = cp.Variable(nvar)

    # Objective function
    obj = _model_objective(A, c, y, objective, regularization, h_epsilon,
                           quantile, reg_l1, reg_l2)

    # Constraints
    constraints = []
    if n_bins > 1 and continuous:
        S = matrix_S(x, splits, order)
        constraints.append(S @ c == 0)

    if monotonic_trend:
        mono_cons = monotonic_trend_constraints(monotonic_trend, c, D, t)
        if isinstance(mono_cons, list):
            constraints.extend(mono_cons)
        else:
            constraints.append(mono_cons)

    if lb is not None:
        constraints.append(A @ c >= lb)
    if ub is not None:
        constraints.append(A @ c <= ub)

    # Solve
    prob = cp.Problem(obj, constraints)

    if solver in ("auto", "ecos", "scs"):
        if solver in ("auto", "ecos"):
            solve_options = {'solver': cp.ECOS, 'verbose': verbose}
        else:
            solve_options = {'solver': cp.SCS, 'verbose': verbose}
    elif solver == "highs":
        solve_options = {'solver': cp.SCIPY}

    if solver == "highs":
        prob.solve(**solve_options, scipy_options={'method': "highs"})
    else:
        prob.solve(**solve_options)

    size_metrics = cp.problems.problem.SizeMetrics(prob)
    status = prob.status
    info = problem_info(status, size_metrics)

    return c.value.reshape((n_bins, order)), info


def socp_separated(x, y, splits, degree, lb, ub, objective,
                   monotonic_trend, h_epsilon, quantile, solver, verbose):

    if solver in ("ecos", "scs"):
        if solver == "ecos":
            solve_options = {'solver': cp.ECOS, 'verbose': verbose}
        else:
            solve_options = {'solver': cp.SCS, 'verbose': verbose}
    elif solver == "highs":
        solve_options = {'solver': cp.SCIPY}

    order = degree + 1
    n_bins = len(splits) + 1

    indices = np.searchsorted(splits, x, side='right')

    c = np.zeros((n_bins, order))

    infos = []
    for i in range(n_bins):
        mask = (indices == i)
        xi = x[mask]
        yi = y[mask]
        ni = len(xi)

        if monotonic_trend in ("ascending", "descending"):
            if order == 2:
                Ai = submatrix_A(ni, xi, order)
                Di = submatrix_D(order)
            else:
                Ai, Di = submatrix_A_D(ni, xi, order)
        else:
            Ai = submatrix_A(ni, xi, order)
            Di = None

        # Decision variables
        ci = cp.Variable(order)

        # Objective function
        obj = _model_objective(Ai, ci, yi, objective, None, h_epsilon,
                               quantile, None, None)

        # Constraints
        constraints = []
        if monotonic_trend:
            mono_cons = monotonic_trend_constraints(monotonic_trend, ci, Di)
            constraints.append(mono_cons)

        if lb is not None:
            constraints.append(Ai @ ci >= lb)
        if ub is not None:
            constraints.append(Ai @ ci <= ub)

        prob = cp.Problem(obj, constraints)

        if solver == "highs":
            prob.solve(**solve_options, scipy_options={'method': "highs"})
        else:
            prob.solve(**solve_options)

        size_metrics = cp.problems.problem.SizeMetrics(prob)
        status = prob.status
        info = problem_info(status, size_metrics)
        infos.append(info)

        c[i, :] = ci.value

    return c, infos



"""
Auxiliary functions for cvxpy formulations.
"""

# Guillermo Navas-Palencia <g.navas.palencia@gmail.com>
# Copyright (C) 2020

import numpy as np


def monotonic_trend_constraints(monotonic_trend, c, D, t=None):
    if monotonic_trend in ("ascending", "convex"):
        return D @ c >= 0
    elif monotonic_trend in ("descending", "concave"):
        return D @ c <= 0
    elif monotonic_trend == "valley":
        return [D[:t, :] @ c <= 0, D[t:, :] @ c >= 0]
    elif monotonic_trend == "peak":
        return [D[:t, :] @ c >= 0, D[t:, :] @ c <= 0]


def compute_change_point(x, y, splits, order, monotonic_trend):
    n_splits = len(splits)
    n_bins = n_splits + 1
    indices = np.searchsorted(splits, x, side='right')

    mean = [y[indices == i].mean() for i in range(n_bins)]
    if monotonic_trend == "peak":
        change_point = np.argmax(mean)
    else:
        change_point = np.argmin(mean)

    if order > 2:
        if change_point >= n_splits:
            change_point = n_splits - 1
        change_point = np.searchsorted(x, splits[change_point], side='right')

    return change_point + 1


def problem_info(status, size_metrics):
    n_variables = size_metrics.num_scalar_variables
    n_constraints = size_metrics.num_scalar_eq_constr
    n_constraints += size_metrics.num_scalar_leq_constr

    info = {
        "status": status,
        "stats": {"n_variables": n_variables, "n_constraints": n_constraints}
    }

    return info



#!/usr/bin/env python

import os
import sys

from setuptools import find_packages, setup, Command
from setuptools.command.test import test as TestCommand


long_description = '''
**RoPWR** is a library written in Python implementing several mathematical
programming formulations to compute the optimal continuous/discontinuous
piecewise polynomial regression given a list of split points.

Read the documentation at: http://gnpalencia.org/ropwr/

RoPWR is distributed under the Apache Software License (Apache 2.0).
'''


class CleanCommand(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        os.system('rm -vrf ./build ./dist ./*.pyc ./*.tgz ./*.egg-info')


# test suites
class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = []

    def run_tests(self):
        # import here, because outside the eggs aren't loaded
        import pytest
        errcode = pytest.main(self.test_args)
        sys.exit(errcode)


# install requirements
install_requires = [
    'cvxpy>=1.1.14',
    'numpy>=1.16',
    'scikit-learn>=0.22',
    'scipy>=1.6.1'
]

# test requirements
tests_require = [
    'pytest',
    'coverage'
]


# Read version file
version_info = {}
with open("ropwr/_version.py") as f:
    exec(f.read(), version_info)


setup(
    name="ropwr",
    version=version_info['__version__'],
    description="RoPWR: Robust Piecewise Regression",
    long_description=long_description,
    author="Guillermo Navas-Palencia",
    author_email="g.navas.palencia@gmail.com",
    packages=find_packages(),
    platforms="any",
    include_package_data=True,
    license="Apache Licence 2.0",
    url="https://github.com/guillermo-navas-palencia/ropwr",
    cmdclass={'clean': CleanCommand, 'test': PyTest},
    python_requires='>=3.7',
    install_requires=install_requires,
    tests_require=tests_require,
    classifiers=[
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3']
    )



"""
Direct method for piecewise regression with l2-norm objective and no other
constraints.
"""

# Guillermo Navas-Palencia <g.navas.palencia@gmail.com>
# Copyright (C) 2020

import numpy as np

from .matrices import submatrix_A


def lsq_direct(x, y, splits, degree):
    order = degree + 1
    n = len(x)
    n_splits = len(splits)
    n_bins = n_splits + 1

    nA = n_bins * order

    A = np.zeros((n, nA))
    S = np.zeros((n_splits, nA))

    indices = np.searchsorted(splits, x, side='right')

    cn = 0
    for i in range(n_bins):
        xi = x[indices == i]
        ni = len(xi)

        pxi = np.ones(ni)
        for j in range(order * i, order * (i + 1)):
            A[cn: cn + ni, j] = pxi
            pxi *= xi

        cn += ni

    exporder = np.arange(order)
    for i, s in enumerate(splits):
        r = np.power(s, exporder)
        S[i, order * i: order * (i + 1)] = r
        S[i, order * (i + 1): order * (i + 2)] = -r

    nM = nA + n_splits
    M = np.zeros((nM, nM))

    M[:nA, :nA] = 2 * A.T.dot(A)
    M[:nA, nA:] = S.T
    M[nA:, :nA] = S

    d = np.zeros(nM)
    d[:nA] = 2 * A.T.dot(y)

    c = np.linalg.solve(M, d)

    info = {
        "status": "optimal",
        "stats": {"n_variables": nA, "n_constraints": nM}
    }

    return c[:nA].reshape((n_bins, order)), info


def lsq_direct_separated(x, y, splits, degree):
    order = degree + 1
    n_bins = len(splits) + 1

    indices = np.searchsorted(splits, x, side='right')

    c = np.zeros((n_bins, order))

    infos = []
    for i in range(n_bins):
        mask = (indices == i)
        xi = x[mask]
        yi = y[mask]
        ni = len(xi)
        Ai = submatrix_A(ni, xi, order)

        ci, _, _, _ = np.linalg.lstsq(Ai, yi, rcond=None)

        info = {
            "status": "optimal",
            "stats": {"n_variables": order, "n_constraints": 0}
        }
        infos.append(info)

        c[i, :] = ci

    return c, infos
